"""
Manages configuration of ``pydistcheck` CLI, including:

  * validating configuration values
  * updating configuuration from files
"""
import os
from dataclasses import dataclass
from typing import Any, Dict

from pydistcheck._compat import tomllib

# putting this in a module-level set to save on the cost of re-computing it inside methods
# in `_Config` that validate configuration.
#
# unit tests confirm that it matches the `_Config` class, so it shouldn't ever drift from that class
_ALLOWED_CONFIG_VALUES = {
    "ignore",
    "inspect",
    "max_allowed_files",
    "max_allowed_size_compressed",
    "max_allowed_size_uncompressed",
    "unexpected_directory_patterns",
    "unexpected_file_patterns",
}

_UNEXPECTED_DIRECTORIES = [
    "*/.appveyor",
    "*/.binder",
    "*/.circleci",
    "*/.git",
    "*/.github",
    "*/.idea",
    "*/.pytest_cache",
    "*/.mypy_cache",
]

_UNEXPECTED_FILES = [
    "*/appveyor.yml",
    "*/.appveyor.yml",
    "*/azure-pipelines.yml",
    "*/.azure-pipelines.yml",
    "*/.cirrus.star",
    "*/.cirrus.yml",
    "*/codecov.yml",
    "*/.codecov.yml",
    "*/.DS_Store",
    "*/.gitignore",
    "*/.gitpod.yml",
    "*/.hadolint.yaml",
    "*/.readthedocs.yaml",
    "*/.travis.yml",
    "*/vsts-ci.yml",
    "*/.vsts-ci.yml",
]


@dataclass
class _Config:
    ignore: str = ""
    inspect: bool = False
    max_allowed_files: int = 2000
    max_allowed_size_compressed: str = "50M"
    max_allowed_size_uncompressed: str = "75M"
    unexpected_directory_patterns: str = ",".join(_UNEXPECTED_DIRECTORIES)
    unexpected_file_patterns: str = ",".join(_UNEXPECTED_FILES)

    def __setattr__(self, name: str, value: Any) -> None:
        attr_name = name.replace("-", "_")
        if attr_name not in _ALLOWED_CONFIG_VALUES:
            raise ValueError(f"Configuration value '{name}' is not recognized by pydistcheck")
        object.__setattr__(self, attr_name, value)

    def update_from_dict(self, input_dict: Dict[str, Any]) -> "_Config":
        for k, v in input_dict.items():
            setattr(self, k, v)
        return self

    def update_from_toml(self, toml_file: str) -> "_Config":
        if not os.path.exists(toml_file):
            return self

        tool_options: Dict[str, Any] = {}
        with open(toml_file, "rb") as f:
            config_dict = tomllib.load(f)
            tool_options = config_dict.get("tool", {}).get("pydistcheck", {})

        # list-like stuff in TOML is expected to be a comma-delimited string when passed as
        # a command-line argument
        patch_dict: Dict[str, Any] = {}
        for k, v in tool_options.items():
            if isinstance(v, list):
                patch_dict[k] = ",".join(v)
        tool_options.update(patch_dict)
        self.update_from_dict(tool_options)
        return self



"""
Code that prints diagnostic information about a distribution.
"""

from pydistcheck.distribution_summary import _DistributionSummary
from pydistcheck.utils import _FileSize


def inspect_distribution(filepath: str) -> None:
    summary = _DistributionSummary.from_file(filename=filepath)
    print("file size")
    compressed_size = _FileSize(summary.compressed_size_bytes, "B")
    uncompressed_size = _FileSize(summary.uncompressed_size_bytes, "B")
    print(f"  * compressed size: {compressed_size}")
    print(f"  * uncompressed size: {uncompressed_size}")
    space_saving = 1.0 - (compressed_size.total_size_bytes / uncompressed_size.total_size_bytes)
    print(f"  * compression space saving: {round(100 * space_saving, 1)}%")

    print("contents")
    print(f"  * directories: {summary.num_directories}")
    print(f"  * files: {summary.num_files} ({len(summary.compiled_objects)} compiled)")

    print("size by extension")
    for extension, size in summary.size_by_file_extension.items():
        size_pct = size / summary.uncompressed_size_bytes
        print(f"  * {extension} - {round(size / 1024.0, 1)}K ({round(size_pct * 100, 1)}%)")



# pylint: disable=unused-import

"""
Central location for weird import stuff used to make the project compatible
with a wide range of dependency versions.
"""

try:
    import tomllib  # type: ignore  # noqa: F401
except ModuleNotFoundError:
    import tomli as tomllib  # type: ignore  # noqa: F401



import os
import uuid
from copy import deepcopy
from dataclasses import fields

import pytest

from pydistcheck.config import _ALLOWED_CONFIG_VALUES, _Config


@pytest.fixture(scope="function")
def base_config():
    return _Config(
        inspect=False,
        max_allowed_files=7,
        max_allowed_size_compressed="1G",
        max_allowed_size_uncompressed="18K",
    )


def test_allowed_config_values_constant_matches_class():
    fields_from_class = {f.name for f in fields(_Config)}
    assert fields_from_class == _ALLOWED_CONFIG_VALUES


def test_config_fixture_works_as_expected(base_config):
    assert base_config.inspect is False
    assert base_config.max_allowed_files == 7
    assert base_config.max_allowed_size_compressed == "1G"
    assert base_config.max_allowed_size_uncompressed == "18K"


def test_config_can_be_initialized_without_any_arguments():
    config = _Config()
    assert config.inspect is False
    assert config.max_allowed_files == 2000
    assert config.max_allowed_size_compressed == "50M"
    assert config.max_allowed_size_uncompressed == "75M"


def test_update_from_dict_raises_exception_for_first_bad_value_encountered(base_config):
    with pytest.raises(
        ValueError, match=r"Configuration value 'a' is not recognized by pydistcheck"
    ):
        base_config.update_from_dict({"a": 7, "b": 8, "inspect": False})


def test_update_from_dict_works_even_if_dict_does_not_include_all_config_values(base_config):
    assert base_config.inspect is False
    base_config.update_from_dict({"inspect": True})
    assert base_config.inspect is True


def test_update_from_dict_works_when_changing_all_values(base_config):
    assert base_config.inspect is False
    assert base_config.max_allowed_files == 7
    assert base_config.max_allowed_size_compressed == "1G"
    assert base_config.max_allowed_size_uncompressed == "18K"
    patch_dict = {
        "ignore": "path-contains-spaces,too-many-files",
        "inspect": True,
        "max_allowed_files": 8,
        "max_allowed_size_compressed": "2G",
        "max_allowed_size_uncompressed": "141K",
        "unexpected_directory_patterns": "*/tests",
        "unexpected_file_patterns": "*.xlsx,data/*.csv",
    }
    assert set(patch_dict.keys()) == _ALLOWED_CONFIG_VALUES, "this test needs to be updated"
    base_config.update_from_dict(patch_dict)
    assert base_config.inspect is True
    assert base_config.max_allowed_files == 8
    assert base_config.max_allowed_size_compressed == "2G"
    assert base_config.max_allowed_size_uncompressed == "141K"
    assert base_config.unexpected_directory_patterns == "*/tests"
    assert base_config.unexpected_file_patterns == "*.xlsx,data/*.csv"


def test_update_from_toml_silently_returns_self_if_file_does_not_exist(base_config):
    original_config = deepcopy(base_config)
    out = base_config.update_from_toml(toml_file=f"{str(uuid.uuid4())}.toml")
    assert out == original_config


def test_update_from_toml_works_for_files_with_no_pydistcheck_configuration(base_config, tmpdir):
    original_config = deepcopy(base_config)
    temp_file = os.path.join(tmpdir, f"{str(uuid.uuid4())}.toml")
    with open(temp_file, "w") as f:
        f.write("""[tool.pylint]\n""")
    base_config.update_from_toml(toml_file=temp_file)
    assert base_config == original_config


def test_update_from_toml_works_for_files_with_empty_pydistcheck_configuration(base_config, tmpdir):
    original_config = deepcopy(base_config)
    temp_file = os.path.join(tmpdir, f"{str(uuid.uuid4())}.toml")
    with open(temp_file, "w") as f:
        f.write("""\n[tool.pylint]\n[tool.pydistcheck]\n""")
    base_config.update_from_toml(toml_file=temp_file)
    assert base_config == original_config


@pytest.mark.parametrize(
    "config_key", ["max_allowed_size_compressed", "max-allowed-size_compressed"]
)
def test_update_from_toml_works_with_underscores_and_hyphens(base_config, tmpdir, config_key):
    temp_file = os.path.join(tmpdir, f"{str(uuid.uuid4())}.toml")
    with open(temp_file, "w") as f:
        f.write(f"[tool.pylint]\n[tool.pydistcheck]\n{config_key} = '2.5G'\n")
    base_config.update_from_toml(toml_file=temp_file)
    assert base_config.max_allowed_size_compressed == "2.5G"


@pytest.mark.parametrize("use_hyphens", [True, False])
def test_update_from_toml_works_with_all_config_values(base_config, tmpdir, use_hyphens):
    temp_file = os.path.join(tmpdir, f"{str(uuid.uuid4())}.toml")
    patch_dict = {
        "ignore": "[\n'path-contains-spaces',\n'too-many-files'\n]",
        "inspect": "true",
        "max_allowed_files": 8,
        "max_allowed_size_compressed": "'3G'",
        "max_allowed_size_uncompressed": "'4.12G'",
        "unexpected_directory_patterns": "[\n'tests/*'\n]",
        "unexpected_file_patterns": "[\n'*.pq',\n'*/tests/data/*.csv']",
    }
    assert set(patch_dict.keys()) == _ALLOWED_CONFIG_VALUES, "this test needs to be updated"
    if use_hyphens:
        patch_dict = {k.replace("_", "-"): v for k, v in patch_dict.items()}
    with open(temp_file, "w") as f:
        lines = ["[tool.pylint]", "[tool.pydistcheck]"] + [
            f"{k} = {v}" for k, v in patch_dict.items()
        ]
        f.write("\n".join(lines))
    base_config.update_from_toml(toml_file=temp_file)
    assert base_config.ignore == "path-contains-spaces,too-many-files"
    assert base_config.inspect is True
    assert base_config.max_allowed_files == 8
    assert base_config.max_allowed_size_compressed == "3G"
    assert base_config.max_allowed_size_uncompressed == "4.12G"
    assert base_config.unexpected_directory_patterns == "tests/*"
    assert base_config.unexpected_file_patterns == "*.pq,*/tests/data/*.csv"


def test_update_from_toml_converts_lists_to_comma_delimited_string(base_config, tmpdir):
    temp_file = os.path.join(tmpdir, f"{str(uuid.uuid4())}.toml")
    with open(temp_file, "w") as f:
        f.write(
            "[tool.pylint]\n[tool.pydistcheck]\n"
            "max_allowed_size_compressed = [\n'2.5G',\n'1.56K',\n'4.236B'\n]\n"
        )
    base_config.update_from_toml(toml_file=temp_file)
    assert base_config.max_allowed_size_compressed == "2.5G,1.56K,4.236B"



import pytest

from pydistcheck.utils import _FileSize


def test_file_size_minimally_works():
    fs = _FileSize(num=1.0, unit_str="G")
    assert fs.total_size_bytes == int(1024**3)


def test_file_size_comparisons_work():
    fs_5mb = _FileSize(num=5.6, unit_str="M")
    fs_6mb = _FileSize(num=4 * (1024**3), unit_str="B")

    assert fs_5mb == fs_5mb
    assert fs_5mb != fs_6mb
    assert fs_5mb != fs_6mb
    assert fs_5mb < fs_6mb
    assert fs_5mb <= fs_6mb
    assert fs_6mb >= fs_5mb
    assert fs_6mb > fs_5mb


@pytest.mark.parametrize(
    "file_size",
    [
        _FileSize.from_number(3 * 1024**2),
        _FileSize.from_string("3M"),
        _FileSize.from_string("3.0M"),
        _FileSize.from_string("3.0000M"),
    ],
)
def test_file_size_from_different_inputs_all_parsed_consistently(file_size):
    assert file_size == _FileSize(num=3.0, unit_str="M")


def test_file_size_from_number_switches_unit_str_based_on_size():
    assert _FileSize.from_number(1.1) == _FileSize(num=1.1, unit_str="B")
    # fractional bytes don't make sense here, so some rounding happens
    # e.g., 0.1 KB is technically 102.4 bytes, which gets rounded to 102
    assert _FileSize.from_number(102) == _FileSize(num=0.1, unit_str="K")
    # 3.456789 * 10**3 = 3711698926.043136
    assert _FileSize.from_number(3711698926) == _FileSize(num=3.456789, unit_str="G")


@pytest.mark.parametrize(
    "file_size,expected_str",
    [
        (_FileSize.from_number(110), "0.1K"),
        (_FileSize.from_number(150), "0.1K"),
        (_FileSize.from_number(1024), "1.0K"),
        (_FileSize.from_number(100 * 1024), "100.0K"),
        (_FileSize.from_number(200 * 1024), "0.2M"),
        (_FileSize.from_number(1024**2), "1.0M"),
        (_FileSize.from_number(400 * 1024**2), "0.4G"),
        (_FileSize.from_number(405 * 1024**2), "0.4G"),
        (_FileSize.from_number(1024**3), "1.0G"),
        (_FileSize.from_number(int(7.5 * 1024**3)), "7.5G"),
    ],
)
def test_file_size_string_representation_looks_correct(file_size, expected_str):
    assert str(file_size) == expected_str



"""
CLI entrypoints
"""

import sys
from typing import List

import click

from pydistcheck.checks import (
    ALL_CHECKS,
    _CompiledObjectsDebugSymbolCheck,
    _DistroTooLargeCompressedCheck,
    _DistroTooLargeUnCompressedCheck,
    _FileCountCheck,
    _FilesOnlyDifferByCaseCheck,
    _MixedFileExtensionCheck,
    _NonAsciiCharacterCheck,
    _SpacesInPathCheck,
    _UnexpectedFilesCheck,
)
from pydistcheck.config import _Config
from pydistcheck.distribution_summary import _DistributionSummary
from pydistcheck.inspect import inspect_distribution
from pydistcheck.utils import _FileSize


@click.command()
@click.argument(
    "filepaths",
    type=click.Path(exists=True),
    nargs=-1,
)
@click.option(
    "--ignore",
    type=str,
    default=_Config.ignore,
    help=(
        "comma-separated list of checks to skip, e.g. "
        "``distro-too-large-compressed,path-contains-spaces``."
    ),
)
@click.option(
    "--inspect",
    is_flag=True,
    show_default=False,
    default=_Config.inspect,
    help="print diagnostic information about the distribution",
)
@click.option(
    "--max-allowed-files",
    default=_Config.max_allowed_files,
    show_default=True,
    type=int,
    help="maximum number of files allowed in the distribution",
)
@click.option(
    "--max-allowed-size-compressed",
    default=_Config.max_allowed_size_compressed,
    show_default=True,
    type=str,
    help=(
        "maximum allowed compressed size, a string like '1.5M' indicating"
        " '1.5 megabytes'. Supported units:\n"
        "  - B = bytes\n"
        "  - K = kilobytes\n"
        "  - M = megabytes\n"
        "  - G = gigabytes"
    ),
)
@click.option(
    "--max-allowed-size-uncompressed",
    default=_Config.max_allowed_size_uncompressed,
    show_default=True,
    type=str,
    help=(
        "maximum allowed uncompressed size, a string like '1.5M' indicating"
        " '1.5 megabytes'. Supported units:\n"
        "  - B = bytes\n"
        "  - K = kilobytes\n"
        "  - M = megabytes\n"
        "  - G = gigabytes"
    ),
)
@click.option(
    "--unexpected-directory-patterns",
    default=_Config.unexpected_directory_patterns,
    show_default=True,
    type=str,
    help=(
        "comma-delimited list of patterns matching directories that are not expected "
        "to be found in the distribution. Patterns should be in the format understood "
        "by ``fnmatch.fnmatchcase()``. See https://docs.python.org/3/library/fnmatch.html."
    ),
)
@click.option(
    "--unexpected-file-patterns",
    default=_Config.unexpected_file_patterns,
    show_default=True,
    type=str,
    help=(
        "comma-delimited list of patterns matching files that are not expected "
        "to be found in the distribution. Patterns should be in the format understood "
        "by ``fnmatch.fnmatchcase()``. See https://docs.python.org/3/library/fnmatch.html."
    ),
)
def check(  # pylint: disable=too-many-arguments
    filepaths: str,
    ignore: str,
    inspect: bool,
    max_allowed_files: int,
    max_allowed_size_compressed: str,
    max_allowed_size_uncompressed: str,
    unexpected_directory_patterns: str,
    unexpected_file_patterns: str,
) -> None:
    """
    Run the contents of a distribution through a set of checks, and warn about
    any problematic characteristics that are detected.
    """
    print("==================== running pydistcheck ====================")
    filepaths_to_check = [click.format_filename(f) for f in filepaths]
    config = _Config()
    kwargs = {
        "ignore": ignore,
        "inspect": inspect,
        "max_allowed_files": max_allowed_files,
        "max_allowed_size_compressed": max_allowed_size_compressed,
        "max_allowed_size_uncompressed": max_allowed_size_uncompressed,
        "unexpected_directory_patterns": unexpected_directory_patterns,
        "unexpected_file_patterns": unexpected_file_patterns,
    }
    kwargs_that_differ_from_defaults = {}
    for k, v in kwargs.items():
        if v != getattr(config, k):
            kwargs_that_differ_from_defaults[k] = v
    config.update_from_toml(toml_file="pyproject.toml")
    config.update_from_dict(input_dict=kwargs_that_differ_from_defaults)

    checks_to_ignore = {x for x in config.ignore.split(",") if x.strip() != ""}
    unrecognized_checks = checks_to_ignore - ALL_CHECKS
    if unrecognized_checks:
        # converting to list + sorting here so outputs are deterministic
        # (since sets don't guarantee ordering)
        error_str = ",".join(sorted(unrecognized_checks))
        print(f"ERROR: found the following unrecognized checks passed via '--ignore': {error_str}")
        sys.exit(1)

    checks = [
        _CompiledObjectsDebugSymbolCheck(),
        _DistroTooLargeCompressedCheck(
            max_allowed_size_bytes=_FileSize.from_string(
                size_str=config.max_allowed_size_compressed
            ).total_size_bytes
        ),
        _DistroTooLargeUnCompressedCheck(
            max_allowed_size_bytes=_FileSize.from_string(
                size_str=config.max_allowed_size_uncompressed
            ).total_size_bytes
        ),
        _FileCountCheck(max_allowed_files=config.max_allowed_files),
        _FilesOnlyDifferByCaseCheck(),
        _MixedFileExtensionCheck(),
        _SpacesInPathCheck(),
        _UnexpectedFilesCheck(
            unexpected_directory_patterns=unexpected_directory_patterns.split(","),
            unexpected_file_patterns=unexpected_file_patterns.split(","),
        ),
        _NonAsciiCharacterCheck(),
    ]

    # filter out ignored checks
    checks = [c for c in checks if c.check_name not in checks_to_ignore]

    any_errors_found = False
    for filepath in filepaths_to_check:
        print(f"\nchecking '{filepath}'")

        if config.inspect:
            print("----- package inspection summary -----")
            inspect_distribution(filepath=filepath)

        print("------------ check results -----------")
        summary = _DistributionSummary.from_file(filename=filepath)
        errors: List[str] = []
        for this_check in checks:
            errors += this_check(distro_summary=summary)

        for i, error_msg in enumerate(sorted(errors)):
            print(f"{i + 1}. {error_msg}")

        num_errors_for_this_file = len(errors)
        if num_errors_for_this_file:
            any_errors_found = True

        print(f"errors found while checking: {num_errors_for_this_file}")

    print("\n==================== done running pydistcheck ===============")

    # now that all files have been checked, be sure to exit with a non-0 code
    # if any errors were found
    sys.exit(int(any_errors_found))



import pydistcheck.checks


def test_all_checks_constant_contains_names_of_all_checks():
    check_classes = []
    for obj_name in dir(pydistcheck.checks):
        obj = getattr(pydistcheck.checks, obj_name)
        if hasattr(obj, "check_name") and obj is not pydistcheck.checks._CheckProtocol:
            check_classes.append(obj.check_name)

    # every check should be in ALL_CHECKS
    assert set(check_classes) == pydistcheck.checks.ALL_CHECKS

    # check class names should be unique
    assert len(pydistcheck.checks.ALL_CHECKS) == len(check_classes)



"""
functions used to analyze compiled objects
"""

import os
import re
import subprocess
import zipfile
from tempfile import TemporaryDirectory
from typing import List, Tuple

from pydistcheck.distribution_summary import _FileInfo

_COMMAND_FAILED = "__command_failed__"
_NO_DEBUG_SYMBOLS = "__no_debug_symbols_found__"
_TOOL_NOT_AVAILABLE = "__tool_not_available__"


def _run_command(args: List[str]) -> str:
    try:
        stdout = subprocess.run(args, capture_output=True, check=True).stdout
        return stdout.decode("utf-8")
    except subprocess.CalledProcessError:
        return _COMMAND_FAILED
    except FileNotFoundError:
        return _TOOL_NOT_AVAILABLE


# commands to dump symbol information, and regular expressions which, if they match
# any lines in the output, indicate that debug symbols have been found
# fmt: off
_COMMANDS_TO_PATTERNS = [
    (["dsymutil", "-s"],                           r"\(N_OSO[\t ]+\)"),
    (["objdump", "--all-headers"],                 r"[\t ]+\.debug_line[\t ]+"),
    (["objdump", "--macho", "--all-headers"],      r"[\t ]+\.debug_line[\t ]+"),
    (["objdump", "-W"],                            r"^Contents of the \.debug"),
    (["objdump", "--macho", "-W"],                 r"^Contents of the \.debug"),
    (["objdump", "-g"],                            r"^Contents of the \.debug"),
    (["objdump", "--macho", "-g"],                 r"^Contents of the \.debug"),
    (["llvm-objdump", "--all-headers"],            r"[\t ]+\.debug_line[\t ]+"),
    (["llvm-objdump", "--macho", "--all-headers"], r"[\t ]+\.debug_line[\t ]+"),
    (["llvm-objdump", "-W"],                       r"^Contents of the \.debug"),
    (["llvm-objdump", "--macho", "-W"],            r"^Contents of the \.debug"),
    (["llvm-objdump", "-g"],                       r"^Contents of the \.debug"),
    (["llvm-objdump", "--macho", "-g"],            r"^Contents of the \.debug"),
    (["readelf", "-S"],                            r"[\t ]+\.debug_[a-z]+[\t ]+")
]
# fmt: on


def _look_for_debug_symbols(lib_file: str) -> Tuple[bool, str]:
    for cmd_args, pattern in _COMMANDS_TO_PATTERNS:
        stdout = _run_command(args=[*cmd_args, lib_file])
        contains_debug_symbols = any(bool(re.search(pattern, x)) for x in stdout.split("\n"))
        if contains_debug_symbols:
            return True, " ".join(cmd_args)
    # if you get here, no debug symbols were found by any tools
    return False, _NO_DEBUG_SYMBOLS


def _get_symbols(cmd_args: List[str], lib_file: str) -> str:
    syms = _run_command(args=[*cmd_args, lib_file])
    return "\n".join([line for line in syms.split("\n") if not (" a " in line or "\ta\t" in line)])


def _nm_reports_debug_symbols(tool_name: str, lib_file: str) -> Tuple[bool, str]:
    exported_symbols = _get_symbols(cmd_args=[tool_name], lib_file=lib_file)
    all_symbols = _get_symbols(cmd_args=[tool_name, "--debug-syms"], lib_file=lib_file)
    return exported_symbols != all_symbols, f"{tool_name} --debug-syms"


def _archive_member_has_debug_symbols(archive_file: str, file_info: _FileInfo) -> Tuple[bool, str]:
    with TemporaryDirectory() as tmp_dir:
        with zipfile.ZipFile(archive_file, mode="r") as zf:
            zf.extractall(path=tmp_dir, members=[file_info.name])
        full_path = os.path.join(tmp_dir, file_info.name)

        # test with tools that produce debug symbols that can be matched with a regex
        has_debug_symbols, cmd_str = _look_for_debug_symbols(lib_file=full_path)
        if has_debug_symbols:
            return True, cmd_str

        # "nm"'s test is like "check if the output is different when a flag is supplied",
        # instead of "test if this tool produces output matching this regex",
        # so it runs separately here
        for nm_tool in ["nm", "llvm-nm"]:
            has_debug_symbols, cmd_str = _nm_reports_debug_symbols(
                tool_name=nm_tool,
                lib_file=full_path,
            )
            if has_debug_symbols:
                return True, cmd_str

    # at this point, none of the checks found debug symbols
    return False, cmd_str



"""
internal-only classes used to manage information about
source distributions and their contents
"""

import os
import pathlib
import tarfile
import zipfile
from collections import OrderedDict, defaultdict
from dataclasses import dataclass
from functools import cached_property
from typing import Dict, List, Tuple, Union


@dataclass
class _DirectoryInfo:
    name: str


@dataclass
class _FileInfo:
    name: str
    file_format: str
    file_extension: str
    is_compiled: bool
    uncompressed_size_bytes: int

    @classmethod
    def from_tarfile_member(
        cls, archive_file: tarfile.TarFile, tar_info: tarfile.TarInfo
    ) -> "_FileInfo":
        member_name = tar_info.name
        file_format, is_compiled = _guess_archive_member_file_format(
            archive_file=archive_file, member_name=member_name
        )
        return cls(
            name=member_name,
            file_format=file_format,
            file_extension=pathlib.Path(member_name).suffix or "no-extension",
            is_compiled=is_compiled,
            uncompressed_size_bytes=tar_info.size,
        )

    @classmethod
    def from_zipfile_member(
        cls, archive_file: zipfile.ZipFile, zip_info: zipfile.ZipInfo
    ) -> "_FileInfo":
        member_name = zip_info.filename
        file_format, is_compiled = _guess_archive_member_file_format(
            archive_file=archive_file, member_name=member_name
        )
        return cls(
            name=member_name,
            file_format=file_format,
            file_extension=pathlib.Path(member_name).suffix or "no-extension",
            is_compiled=is_compiled,
            uncompressed_size_bytes=zip_info.file_size,
        )


# references:
#   * https://en.wikipedia.org/wiki/List_of_file_signatures
#   * https://github.com/apple-oss-distributions/xnu/blob/5c2921b07a2480ab43ec66f5b9e41cb872bc554f/EXTERNAL_HEADERS/mach-o/loader.h#L65  # noqa: E501
#   * https://github.com/apple-oss-distributions/cctools/blob/658da8c66b4e184458f9c810deca9f6428a773a5/include/mach-o/fat.h#L48  # noqa: E501
#   * https://github.com/matthew-brett/delocate/blob/df86dddd7c94a93b5c03948b8c127ba0777e2a4d/delocate/tools.py#L166  # noqa: E501
#   * https://learn.microsoft.com/en-us/previous-versions/ms809762(v=msdn.10)
#   * https://en.wikipedia.org/wiki/Portable_Executable

# DOS MZ (.dll, .exe)
_DOS_MZ_MAGIC_FIRST_2_BYTES = {
    0x5A4D.to_bytes(2, "little"),
}

# ELF (.so, .o)
_ELF_MAGIC_FIRST_4_BYTES = {
    0x7F454C46.to_bytes(4, "little"),
    0x7F454C46.to_bytes(4, "big"),
}

# MACH-O (.dylib)
_MACH_O_MAGIC_FIRST_4_BYTES = {
    0xFEEDFACE.to_bytes(4, "little"),  # 32-bit
    0xFEEDFACE.to_bytes(4, "big"),  # 32-bit
    0xFEEDFACF.to_bytes(4, "little"),  # 64-bit
    0xFEEDFACF.to_bytes(4, "big"),  # 64-bit
    0xCAFEBABE.to_bytes(4, "big"),  # mach-o fat binary
    0xCAFEBABF.to_bytes(4, "big"),  # mach-o fat binary
}


class _FileFormat:
    ELF = "ELF"
    MACH_O = "Mach-O"
    OTHER = "Other"
    WINDOWS_PE = "Windows PE"


def _guess_archive_member_file_format(
    archive_file: Union[tarfile.TarFile, zipfile.ZipFile], member_name: str
) -> Tuple[str, bool]:
    """
    The approach in this function was inspired by similar code in
    https://github.com/matthew-brett/delocate, so that project's license is included
    in distributions of ``pydistcheck`` at path ``LICENSES/DELOCATE_LICENSE``.

    Returns a two-item tuple of the form ``(file_format, is_compiled)``.
    """
    if isinstance(archive_file, zipfile.ZipFile):
        with archive_file.open(name=member_name, mode="r") as f:
            header = f.read(4)
    else:
        fileobj = archive_file.extractfile(member_name)
        assert fileobj is not None, (
            f"'{member_name}' not found. This is a bug in pydistcheck."
            "Report it at https://github.com/jameslamb/pydistcheck/issues."
        )
        header = fileobj.read(4)

    if header in _ELF_MAGIC_FIRST_4_BYTES:
        return _FileFormat.ELF, True
    if header in _MACH_O_MAGIC_FIRST_4_BYTES:
        return _FileFormat.MACH_O, True
    if header[:2] in _DOS_MZ_MAGIC_FIRST_2_BYTES:
        return _FileFormat.WINDOWS_PE, True

    return _FileFormat.OTHER, False


@dataclass
class _DistributionSummary:
    compressed_size_bytes: int
    directories: List[_DirectoryInfo]
    files: List[_FileInfo]
    original_file: str

    @classmethod
    def from_file(cls, filename: str) -> "_DistributionSummary":
        compressed_size_bytes = os.path.getsize(filename)
        directories: List[_DirectoryInfo] = []
        files: List[_FileInfo] = []
        if filename.endswith("gz"):
            with tarfile.open(filename, mode="r:gz") as tf:
                for tar_info in tf.getmembers():
                    if tar_info.isfile():
                        files.append(
                            _FileInfo.from_tarfile_member(archive_file=tf, tar_info=tar_info)
                        )
                    else:
                        directories.append(_DirectoryInfo(name=tar_info.name))
        else:
            # assume anything else can be opened with zipfile
            with zipfile.ZipFile(filename, mode="r") as f:
                for zip_info in f.infolist():
                    if not zip_info.is_dir():
                        files.append(
                            _FileInfo.from_zipfile_member(archive_file=f, zip_info=zip_info)
                        )
                    else:
                        directories.append(_DirectoryInfo(name=zip_info.filename))

        return cls(
            compressed_size_bytes=compressed_size_bytes,
            directories=directories,
            files=files,
            original_file=filename,
        )

    @property
    def all_paths(self) -> List[str]:
        return self.file_paths + self.directory_paths

    @property
    def compiled_objects(self) -> List[_FileInfo]:
        return [f for f in self.files if f.is_compiled is True]

    @cached_property
    def count_by_file_extension(self) -> Dict[str, int]:
        return {
            file_extension: len(file_list)
            for file_extension, file_list in self.files_by_extension.items()
        }

    @property
    def directory_paths(self) -> List[str]:
        return [d.name for d in self.directories]

    @cached_property
    def files_by_extension(self) -> Dict[str, List[_FileInfo]]:
        out: Dict[str, List[_FileInfo]] = defaultdict(list)
        for f in self.files:
            out[f.file_extension].append(f)
        return out

    @property
    def file_paths(self) -> List[str]:
        return [f.name for f in self.files]

    @property
    def num_directories(self) -> int:
        return len(self.directories)

    @property
    def num_files(self) -> int:
        return len(self.files)

    @property
    def uncompressed_size_bytes(self) -> int:
        return sum(f.uncompressed_size_bytes for f in self.files)

    @property
    def size_by_file_extension(self) -> OrderedDict:
        """
        Aggregate file sizes in a distribution by extension.

        :return: An OrderedDict where keys are file extensions and values are the total size in
                 bytes occupied by such files in the distribution. Sorted in descending
                 order by size.
        """
        summary_dict = {
            file_extension: sum(f.uncompressed_size_bytes for f in files)
            for file_extension, files in self.files_by_extension.items()
        }
        sorted_sizes = list(summary_dict.items())
        sorted_sizes.sort(key=lambda x: x[1], reverse=True)
        out = OrderedDict()
        for file_extension, size_in_bytes in sorted_sizes:
            out[file_extension] = size_in_bytes
        return out



import os
import re
from sys import platform

import pytest
from click.testing import CliRunner, Result

from pydistcheck.cli import check

BASE_PACKAGES = ["base-package-0.1.0.tar.gz", "base-package-0.1.0.zip"]
PROBLEMATIC_PACKAGES = ["problematic-package-0.1.0.tar.gz", "problematic-package-0.1.0.zip"]
MACOS_SUFFIX = "macosx_10_15_x86_64.macosx_11_6_x86_64.macosx_12_5_x86_64.whl"
MANYLINUX_SUFFIX = "manylinux_2_28_x86_64.manylinux_2_5_x86_64.manylinux1_x86_64.whl"
BASE_WHEELS = [
    f"baseballmetrics-0.1.0-py3-none-{MACOS_SUFFIX}",
    f"baseballmetrics-0.1.0-py3-none-{MANYLINUX_SUFFIX}",
]
WHEELS_WITH_DEBUG_SYMBOLS = [
    f"debug-baseballmetrics-0.1.0-py3-none-{MACOS_SUFFIX}",
    f"debug-baseballmetrics-py3-none-{MANYLINUX_SUFFIX}",
]
TEST_DATA_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "data")


def _assert_log_matches_pattern(result: Result, pattern: str, num_times: int = 1) -> None:
    log_lines = result.output.split("\n")
    match_results = [bool(re.search(pattern, log_line)) for log_line in log_lines]
    num_matches_found = sum(match_results)
    msg = (
        f"Expected to find 1 instance of '{pattern}' in logs, "
        f"found {num_matches_found} in {log_lines}."
    )
    assert num_matches_found == num_times, msg


@pytest.mark.parametrize("distro_file", BASE_PACKAGES + BASE_WHEELS)
def test_check_runs_without_error(distro_file):
    runner = CliRunner()
    result = runner.invoke(check, [os.path.join(TEST_DATA_DIR, distro_file)])
    assert result.exit_code == 0


def test_check_fails_with_informative_error_if_file_doesnt_exist():
    runner = CliRunner()
    result = runner.invoke(check, ["some-garbage.exe"])
    assert result.exit_code > 0
    _assert_log_matches_pattern(result, r"Path 'some\-garbage\.exe' does not exist\.$")


def test_check_runs_for_all_files_before_exiting():
    runner = CliRunner()
    result = runner.invoke(
        check,
        [
            "--max-allowed-size-compressed=5B",
            os.path.join(TEST_DATA_DIR, "base-package-0.1.0.tar.gz"),
            os.path.join(TEST_DATA_DIR, "base-package-0.1.0.zip"),
        ],
    )
    assert result.exit_code == 1

    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"^1\. \[distro\-too\-large\-compressed\] Compressed size [0-9]+\.[0-9]+K is "
            r"larger than the allowed size \(5\.0B\)\.$"
        ),
        num_times=2,
    )
    _assert_log_matches_pattern(
        result=result, pattern="errors found while checking\\: 1", num_times=2
    )


@pytest.mark.parametrize("distro_file", BASE_PACKAGES)
def test_check_respects_ignore_with_one_check(distro_file):
    runner = CliRunner()

    # fails when running all checks
    result = runner.invoke(
        check, [os.path.join(TEST_DATA_DIR, distro_file), "--max-allowed-files=1"]
    )
    assert result.exit_code == 1

    # fails with --ignore if the failing check isn't mentioned in --ignore
    result = runner.invoke(
        check,
        [
            os.path.join(TEST_DATA_DIR, distro_file),
            "--ignore=path-contains-spaces",
            "--max-allowed-files=1",
        ],
    )
    assert result.exit_code == 1

    # succeeds if the failing check is mentioned in --ignore
    result = runner.invoke(
        check,
        [
            os.path.join(TEST_DATA_DIR, distro_file),
            "--ignore=too-many-files",
            "--max-allowed-files=1",
        ],
    )
    assert result.exit_code == 0
    _assert_log_matches_pattern(result, "errors found while checking\\: 0")


@pytest.mark.parametrize("distro_file", BASE_PACKAGES)
def test_check_respects_ignore_with_multiple_checks(distro_file):
    runner = CliRunner()

    # fails when running all checks
    result = runner.invoke(
        check,
        [
            os.path.join(TEST_DATA_DIR, distro_file),
            "--max-allowed-files=1",
            "--max-allowed-size-compressed=1B",
        ],
    )
    assert result.exit_code == 1

    # fails with --ignore if all of the failing checks aren't mentioned in --ignore
    result = runner.invoke(
        check,
        [
            os.path.join(TEST_DATA_DIR, distro_file),
            "--ignore=path-contains-spaces,too-many-files",
            "--max-allowed-files=1",
            "--max-allowed-size-compressed=1B",
        ],
    )
    _assert_log_matches_pattern(
        result,
        (
            r"^1\. \[distro\-too\-large\-compressed\] Compressed size [0-9]+\.[0-9]+K is "
            r"larger than the allowed size \(1\.0B\)\.$"
        ),
    )
    assert result.exit_code == 1

    # succeeds if the failing checks are mentioned in --ignore
    result = runner.invoke(
        check,
        [
            os.path.join(TEST_DATA_DIR, distro_file),
            "--ignore=too-many-files,distro-too-large-compressed",
            "--max-allowed-files=1",
        ],
    )
    assert result.exit_code == 0
    _assert_log_matches_pattern(result, "errors found while checking\\: 0")


@pytest.mark.parametrize("distro_file", BASE_PACKAGES)
def test_check_fails_with_expected_error_if_one_check_is_unrecognized(distro_file):
    result = CliRunner().invoke(
        check,
        [os.path.join(TEST_DATA_DIR, distro_file), "--ignore=too-many-files,random-nonsense"],
    )
    assert result.exit_code == 1
    _assert_log_matches_pattern(
        result,
        (
            r"ERROR\: found the following unrecognized checks passed via '\-\-ignore'\: "
            r"random\-nonsense"
        ),
    )


@pytest.mark.parametrize("distro_file", BASE_PACKAGES)
def test_check_fails_with_expected_error_if_multiple_checks_are_unrecognized(distro_file):
    result = CliRunner().invoke(
        check,
        [
            os.path.join(TEST_DATA_DIR, distro_file),
            "--ignore=garbage,too-many-files,random-nonsense,other-trash",
        ],
    )
    assert result.exit_code == 1
    _assert_log_matches_pattern(
        result,
        (
            r"ERROR\: found the following unrecognized checks passed via '\-\-ignore'\: "
            r"garbage,other\-trash,random\-nonsense"
        ),
    )


@pytest.mark.parametrize("distro_file", BASE_PACKAGES)
def test_check_respects_max_allowed_files(distro_file):
    runner = CliRunner()
    result = runner.invoke(
        check, [os.path.join(TEST_DATA_DIR, distro_file), "--max-allowed-files=1"]
    )
    assert result.exit_code == 1

    _assert_log_matches_pattern(
        result, r"^1\. \[too\-many\-files\] Found 11 files\. Only 1 allowed\.$"
    )
    _assert_log_matches_pattern(result, "errors found while checking\\: 1")


@pytest.mark.parametrize(
    "size_str",
    [
        "1K",
        "3.999K",
        "0.0002M",
        "0.00000008G",
        "708B",
    ],
)
@pytest.mark.parametrize("distro_file", BASE_PACKAGES)
def test_check_respects_max_allowed_size_compressed(size_str, distro_file):
    runner = CliRunner()
    result = runner.invoke(
        check,
        [os.path.join(TEST_DATA_DIR, distro_file), f"--max-allowed-size-compressed={size_str}"],
    )
    assert result.exit_code == 1

    _assert_log_matches_pattern(
        result,
        (
            r"^1\. \[distro\-too\-large\-compressed\] Compressed size [0-9]+\.[0-9]+K is "
            r"larger than the allowed size \([0-9]+\.[0-9]+[BKMG]\)\.$"
        ),
    )
    _assert_log_matches_pattern(result, "errors found while checking\\: 1")


@pytest.mark.parametrize(
    "size_str",
    [
        "1K",
        "3.999K",
        "0.0002M",
        "0.00000008G",
        "708B",
    ],
)
@pytest.mark.parametrize("distro_file", BASE_PACKAGES)
def test_check_respects_max_allowed_size_uncompressed(size_str, distro_file):
    runner = CliRunner()
    result = runner.invoke(
        check,
        [os.path.join(TEST_DATA_DIR, distro_file), f"--max-allowed-size-uncompressed={size_str}"],
    )
    assert result.exit_code == 1

    _assert_log_matches_pattern(
        result,
        (
            r"^1\. \[distro\-too\-large\-uncompressed\] Uncompressed size [0-9]+\.[0-9]+K is "
            r"larger than the allowed size \([0-9]+\.[0-9]+[BKMG]\)\.$"
        ),
    )
    _assert_log_matches_pattern(result, "errors found while checking\\: 1")


@pytest.mark.parametrize("distro_file", BASE_PACKAGES)
def test_check_prefers_pyproject_toml_to_defaults(distro_file, tmp_path):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        with open("pyproject.toml", "w") as f:
            f.write("[tool.pylint]\n[tool.pydistcheck]\nmax_allowed_size_uncompressed = '7B'\n")
        result = runner.invoke(
            check,
            [os.path.join(TEST_DATA_DIR, distro_file)],
        )

    assert result.exit_code == 1
    _assert_log_matches_pattern(
        result,
        (
            r"^1\. \[distro\-too\-large\-uncompressed\] Uncompressed size [0-9]+\.[0-9]+K is "
            r"larger than the allowed size \(7\.0B\)\.$"
        ),
    )
    _assert_log_matches_pattern(result, "errors found while checking\\: 1")


@pytest.mark.parametrize("distro_file", BASE_PACKAGES)
def test_check_handles_ignore_list_in_pyproject_toml_correctly(distro_file, tmp_path):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        result = runner.invoke(
            check,
            [os.path.join(TEST_DATA_DIR, distro_file), "--max-allowed-files=1"],
        )
        assert result.exit_code == 1
        _assert_log_matches_pattern(
            result, r"^1\. \[too\-many\-files\] Found [0-9]+ files\. Only 1 allowed\.$"
        )
        with open("pyproject.toml", "w") as f:
            f.write(
                "[tool.pylint]\n[tool.pydistcheck]\n"
                "ignore = [\n'too-many-files', 'path-contains-non-ascii-characters'\n]\n"
            )
        result = runner.invoke(
            check,
            [os.path.join(TEST_DATA_DIR, distro_file), "--max-allowed-files=1"],
        )
        assert result.exit_code == 0
        _assert_log_matches_pattern(result, "errors found while checking\\: 0")


@pytest.mark.parametrize("distro_file", BASE_PACKAGES)
def test_check_prefers_keyword_args_to_pyrpoject_toml_and_defaults(distro_file, tmp_path):
    runner = CliRunner()
    with runner.isolated_filesystem(temp_dir=tmp_path):
        with open("pyproject.toml", "w") as f:
            f.write("[tool.pylint]\n[tool.pydistcheck]\nmax_allowed_size_uncompressed = '7B'\n")
        result = runner.invoke(
            check,
            [os.path.join(TEST_DATA_DIR, distro_file), "--max-allowed-size-uncompressed=123B"],
        )

    assert result.exit_code == 1
    _assert_log_matches_pattern(
        result,
        (
            r"^1\. \[distro\-too\-large\-uncompressed\] Uncompressed size [0-9]+\.[0-9]+K is "
            r"larger than the allowed size \(0\.1K\)\.$"
        ),
    )
    _assert_log_matches_pattern(result, "errors found while checking\\: 1")


@pytest.mark.parametrize("distro_file", PROBLEMATIC_PACKAGES)
def test_files_only_differ_by_case_works(distro_file):
    runner = CliRunner()
    result = runner.invoke(
        check,
        [os.path.join(TEST_DATA_DIR, distro_file)],
    )
    assert result.exit_code == 1

    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"^1\. \[files\-only\-differ\-by\-case\] Found files which differ only by case\. "
            r"Files\: problematic\-package\-0\.1\.0/problematic_package/Question\.py"
            r",problematic\-package\-0\.1\.0/problematic_package/question\.PY"
            r",problematic\-package\-0\.1\.0/problematic_package/question\.py"
        ),
    )
    _assert_log_matches_pattern(result=result, pattern=r"errors found while checking\: [0-9]{1}")


@pytest.mark.parametrize("distro_file", PROBLEMATIC_PACKAGES)
def test_mixed_file_extension_use_works(distro_file):
    runner = CliRunner()
    result = runner.invoke(
        check,
        [os.path.join(TEST_DATA_DIR, distro_file)],
    )
    assert result.exit_code == 1

    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"^2\. \[mixed\-file\-extensions\] Found a mix of file extensions for the "
            r"same file type\: \.NDJSON \(1\), \.jsonl \(1\), \.ndjson \(1\)"
        ),
    )
    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"^3\. \[mixed\-file\-extensions\] Found a mix of file extensions for the "
            r"same file type\: \.yaml \(2\), \.yml \(1\)"
        ),
    )

    _assert_log_matches_pattern(result=result, pattern=r"errors found while checking\: [0-9]{1}")


@pytest.mark.parametrize("distro_file", PROBLEMATIC_PACKAGES)
def test_path_contains_non_ascii_characters_works(distro_file):
    runner = CliRunner()
    result = runner.invoke(
        check,
        [os.path.join(TEST_DATA_DIR, distro_file)],
    )
    assert result.exit_code == 1

    # NOTE: matching a variable number of '?' because zip and tar (and even different
    #       implementations of those) encode non-ASCII characters in paths differently
    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"^4\. \[path\-contains\-non\-ascii\-characters\] Found file path containing non-ASCII "
            r"characters\: "
            r"'problematic\-package\-0\.1\.0/problematic_package/\?+veryone-loves-python\.py'"
        ),
    )

    _assert_log_matches_pattern(result=result, pattern=r"errors found while checking\: [0-9]{1}")


@pytest.mark.parametrize("distro_file", PROBLEMATIC_PACKAGES)
def test_path_contains_spaces_works(distro_file):
    runner = CliRunner()
    result = runner.invoke(
        check,
        [os.path.join(TEST_DATA_DIR, distro_file)],
    )
    assert result.exit_code == 1

    # should report a file with spaces in a directory without spaces
    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"^5\. \[path\-contains\-spaces\] "
            r"Found path with spaces\: 'problematic\-package\-0\.1\.0/beep boop\.ini"
        ),
    )

    # should report a directory with spaces
    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"^6\. \[path\-contains\-spaces\] Found path with spaces\: "
            r"'problematic\-package\-0\.1\.0/problematic_package/bad code[/]*"
        ),
    )

    # should report a file without spaces inside a directory with spaces
    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"^7\. \[path\-contains\-spaces\] Found path with spaces\: "
            r"'problematic\-package\-0\.1\.0/problematic_package/bad code/__init__\.py"
        ),
    )
    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"^8\. \[path\-contains\-spaces\] Found path with spaces\: "
            r"'problematic\-package\-0\.1\.0/problematic_package/bad code/ship\-it\.py"
        ),
    )

    _assert_log_matches_pattern(result=result, pattern=r"errors found while checking\: [0-9]{1}")


@pytest.mark.parametrize("distro_file", PROBLEMATIC_PACKAGES)
def test_unexpected_files_check_works(distro_file):
    runner = CliRunner()
    result = runner.invoke(
        check,
        [os.path.join(TEST_DATA_DIR, distro_file)],
    )
    assert result.exit_code == 1

    # directory
    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"^9\. \[unexpected\-files\] Found unexpected directory "
            r"'problematic\-package\-0\.1\.0/\.git[/]{0,1}'\."
        ),
    )

    # root-level files
    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"^10\. \[unexpected\-files\] Found unexpected file "
            r"'problematic\-package\-0\.1\.0/\.gitignore'\."
        ),
    )
    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"^11\. \[unexpected\-files\] Found unexpected file "
            r"'problematic\-package\-0\.1\.0/\.hadolint\.yaml'\."
        ),
    )

    # nested files
    _assert_log_matches_pattern(
        result=result,
        pattern=(
            r"^12\. \[unexpected\-files\] Found unexpected file "
            r"'problematic\-package\-0\.1\.0/problematic_package/\.gitignore'\."
        ),
    )

    _assert_log_matches_pattern(result=result, pattern=r"errors found while checking\: [0-9]{1}")


@pytest.mark.parametrize("distro_file", PROBLEMATIC_PACKAGES)
def test_cli_raises_exactly_the_expected_number_of_errors_for_the_problematic_package(distro_file):
    runner = CliRunner()
    result = runner.invoke(
        check,
        [os.path.join(TEST_DATA_DIR, distro_file)],
    )
    assert result.exit_code == 1
    _assert_log_matches_pattern(result=result, pattern=r"errors found while checking\: 12$")


@pytest.mark.parametrize("distro_file", WHEELS_WITH_DEBUG_SYMBOLS)
def test_debug_symbols_check_works(distro_file):
    runner = CliRunner()
    result = runner.invoke(
        check,
        [os.path.join(TEST_DATA_DIR, distro_file)],
    )
    assert result.exit_code == 1, result.output
    if "macosx" in distro_file:
        if platform.startswith("cygwin") or platform.startswith("win"):
            debug_cmd = r"'llvm\-nm \-\-debug\-syms \"lib/lib_baseballmetrics\.dylib\"'\."
        else:
            # dsymutil works on both macOS and Linux
            debug_cmd = r"'dsymutil \-s \"lib/lib_baseballmetrics\.dylib\"'\."
    else:
        debug_cmd = r"'objdump \-\-all\-headers \"lib/lib_baseballmetrics\.so\"'\."

    expected_msg = (
        r"^1\. \[compiled\-objects\-have\-debug\-symbols\] "
        r"Found compiled object containing debug symbols\. "
        r"For details, extract the distribution contents and run "
    )
    expected_msg += debug_cmd
    _assert_log_matches_pattern(result=result, pattern=expected_msg)


# --------------------- #
# pydistcheck --inspect #
# --------------------- #


@pytest.mark.parametrize("distro_file", BASE_PACKAGES)
def test_inspect_runs_before_checks(distro_file):
    runner = CliRunner()
    result = runner.invoke(
        check, ["--inspect", "--max-allowed-files=1", os.path.join(TEST_DATA_DIR, distro_file)]
    )
    assert result.exit_code == 1

    _assert_log_matches_pattern(
        result, r"^1\. \[too\-many\-files\] Found 11 files\. Only 1 allowed\.$"
    )
    _assert_log_matches_pattern(result, "errors found while checking\\: 1")
    _assert_log_matches_pattern(result, r"^size by extension$")



"""
miscellaneous helper classes and functions that are
not specific to package distributions
"""

from typing import Any, Tuple

_UNIT_TO_NUM_BYTES = {"B": 1, "K": 1024, "M": 1024**2, "G": 1024**3}


def _recommend_size_str(num_bytes: int) -> Tuple[float, str]:
    if num_bytes < int(0.1 * 1024):
        return float(num_bytes), "B"
    if num_bytes <= (0.1 * 1024**2):
        return float(num_bytes) / 1024.0, "K"
    if num_bytes <= (0.1 * 1024**3):
        return float(num_bytes) / (1024**2), "M"

    return float(num_bytes) / (1024**3), "G"


class _FileSize:
    def __init__(self, num: float, unit_str: str):
        self._num = num
        self._unit_str = unit_str

    @classmethod
    def from_number(cls, num: int) -> "_FileSize":
        num_bytes, unit_str = _recommend_size_str(num_bytes=num)
        return cls(num=num_bytes, unit_str=unit_str)

    @classmethod
    def from_string(cls, size_str: str) -> "_FileSize":
        return cls(num=float(size_str[:-1]), unit_str=size_str[-1])

    @property
    def total_size_bytes(self) -> int:
        return int(self._num * _UNIT_TO_NUM_BYTES[self._unit_str])

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, type(self)) and self.total_size_bytes == other.total_size_bytes

    def __ge__(self, other: "_FileSize") -> bool:
        return self.total_size_bytes >= other.total_size_bytes

    def __gt__(self, other: "_FileSize") -> bool:
        return self.total_size_bytes > other.total_size_bytes

    def __le__(self, other: "_FileSize") -> bool:
        return self.total_size_bytes <= other.total_size_bytes

    def __lt__(self, other: "_FileSize") -> bool:
        return self.total_size_bytes < other.total_size_bytes

    def __ne__(self, other: Any) -> bool:
        return not self == other

    def __str__(self) -> str:
        num_bytes, unit_str = _recommend_size_str(self.total_size_bytes)
        return f"{round(num_bytes, 1)}{unit_str}"
